/*
 * $Id$
 * Copyright (c) 2001 Stefan Matthias Aust.  All Rights Reserved.
 *
 * This file is part of the 3plus4 expression parser.  You are granted the
 * right to use this code in GPL based projects in which case this code is
 * also protected by the GPL, or in other projects as long as you give credit
 * to Stefan Matthias Aust (sma@3plus4.de) and make all modifications or
 * extensions to this code freely available.
 */
package de.bo.expressionparser;

import java.lang.reflect.*;

/**
 * This class is the abstract superclass of all nodes of a parsed expression.
 * It is also used as container to collect all that tiny classes for +, - etc.
 * A node can be evaluated using <code>eval()</code>.
 * @see Parser
 *
 * @author Stefan Matthias Aust <sma@3plus4.de>
 */
public abstract class Node {

    public abstract double eval(double x);

    public static class Number extends Node {
        public Number(double value) { this.value = value; }
        public double eval(double x) { return value; }
        public String toString() { return String.valueOf(value); }
        protected double value;
    }

    public static class Variable extends Node {
        public Variable() { }
        public double eval(double x) { return x; }
        public String toString() { return "x"; }
    }

    public static abstract class Operation extends Node {
        public Operation(Node a, Node b) {
            this.a = a;
            this.b = b;
        }
        public String toString() {
            return "(" + a + op() + b + ")";
        }
        protected abstract String op();
        protected Node a;
        protected Node b;
    }

    public static class Plus extends Operation {
        public Plus(Node a, Node b) { super(a, b); }
        public double eval(double x) { return a.eval(x) + b.eval(x); }
        protected String op() { return "+"; }
    }

    public static class Minus extends Operation {
        public Minus(Node a, Node b) { super(a, b); }
        public double eval(double x) { return a.eval(x) - b.eval(x); }
        protected String op() { return "-"; }
    }

    public static class Times extends Operation {
        public Times(Node a, Node b) { super(a, b); }
        public double eval(double x) { return a.eval(x) * b.eval(x); }
        protected String op() { return "*"; }
    }

    public static class Divide extends Operation {
        public Divide(Node a, Node b) { super(a, b); }
        public double eval(double x) { return a.eval(x) / b.eval(x); }
        protected String op() { return "/"; }
    }

    public static class Power extends Operation {
        public Power(Node a, Node b) { super(a, b); }
        public double eval(double x) { return Math.pow(a.eval(x), b.eval(x)); }
        protected String op() { return "^"; }
    }

    public static class Function extends Node {
        public Function(String name, Node a) throws NoSuchMethodException {
            this.fun = Math.class.getMethod(name, new Class[]{Double.TYPE});
            this.a = a;
        }
        public double eval(double x) {
            try {
                Object[] args = new Object[]{new Double(a.eval(x))};
                return ((Double)fun.invoke(null, args)).doubleValue();
            } catch (InvocationTargetException e) {
                throw (RuntimeException)e.getTargetException();
            } catch (IllegalAccessException e) {
                throw new Error("illegal access to Math." + fun.getName());
            }
        }
        public String toString() {
            return fun.getName() + "(" + a + ")";
        }
        private Method fun;
        private Node a;
    }
}
