/*
 * $Id$
 * Copyright (c) 2001 Stefan Matthias Aust.  All Rights Reserved.
 *
 * This file is part of the 3plus4 expression parser.  You are granted the
 * right to use this code in GPL based projects in which case this code is
 * also protected by the GPL, or in other projects as long as you give credit
 * to Stefan Matthias Aust (sma@3plus4.de) and make all modifications or
 * extensions to this code freely available.
 */
package de.bo.expressionparser;

import java.io.*;

/**
 * This parser understands sentences of the following grammar:
 * <blockquote>
 * expression = term {"+" | "-" term}.<br>
 * term = factor {"*" | "/" factor}.<br>
 * factor = primary {"^" number}.<br>
 * primary = variable | function | number | "(" expression ")".<br>
 * variable = "x".<br>
 * function = name primary.<br>
 * number = ["+" | "-"] digits ["." digits] ["e" digits].<br>
 * digits = "0".."9" {"0".."9"}.<br>
 * </blockquote>
 *
 * @author Stefan Matthias Aust <sma@3plus4.de>
 */
public class Parser {

    public static Node parse(String s) {
        return new Parser(s).parseExpression();
    }

    public Parser(String s) {
        scanner = new Scanner(new StringReader(s));
        advance();
    }

    protected Node parseExpression() {
        Node n = parseTerm();
        while (token == Token.PLUS || token == Token.MINUS) {
            if (token == Token.PLUS) {
                advance();
                n = new Node.Plus(n, parseTerm());
            } else {
                advance();
                n = new Node.Minus(n, parseTerm());
            }
        }
        return n;
    }

    protected Node parseTerm() {
        Node n = parseFactor();
        while (token == Token.TIMES || token == Token.DIVIDE) {
            if (token == Token.TIMES) {
                advance();
                n = new Node.Times(n, parseFactor());
            } else {
                advance();
                n = new Node.Divide(n, parseFactor());
            }
        }
        return n;
    }

    protected Node parseFactor() {
        Node n = parsePrimary();
        while (token == Token.POWER) {
            advance();
            n = new Node.Power(n, parseNumber());
        }
        return n;
    }

    protected Node parsePrimary() {
        if (token == Token.LPAREN) {
            advance();
            Node n = parseExpression();
            if (token != Token.RPAREN)
                throw new Error("missing )");
            advance();
            return n;
        }
        if (token == Token.VARIABLE) {
            advance();
            return new Node.Variable();
        }
        if (token.isName()) {
            return parseFunction();
        }
        return parseNumber();
    }

    protected Node parseFunction() {
        String name = token.getName();
        name.length();
        try {
            advance();
            return new Node.Function(name, parsePrimary());
        } catch (NoSuchMethodException e) {
            throw new Error("unknown function " + name);
        }
    }

    protected Node parseNumber() {
        boolean negative = false;
        if (token == Token.MINUS) {
            advance();
            negative = true;
        } else if (token == Token.PLUS)
            advance();
        if (!token.isNumber())
            //throw new Error("number expected");
            return null;
        double number = token.getNumber();
        if (negative)
            number = -number;
        Node n = new Node.Number(number);
        advance();
        return n;
    }

    protected void advance() {
        try {
            token = scanner.nextToken();
        } catch (IOException e) {
            throw new Error("i/o error");
        }
    }

    private Scanner scanner;
    protected Token token;
}