/*
 * $Id$
 * Copyright (c) 2001 Stefan Matthias Aust.  All Rights Reserved.
 *
 * This file is part of the 3plus4 expression parser.  You are granted the
 * right to use this code in GPL based projects in which case this code is
 * also protected by the GPL, or in other projects as long as you give credit
 * to Stefan Matthias Aust (sma@3plus4.de) and make all modifications or
 * extensions to this code freely available.
 */
package de.bo.expressionparser;

import java.io.Reader;
import java.io.IOException;

/**
 * A scanner breaks the input stream into handy tokens which are then
 * processed by the parser. This scanner knows about mathematic operations
 * like +, -, *, / and ^ as well as named functions, like sin, syntax like
 * parenthesises, variables and floating point numbers.
 *
 * @see Token
 * @see Parser
 *
 * @author Stefan Matthias Aust <sma@3plus4.de>
 */
class Scanner {

    public Scanner(Reader r) {
        this.r = r;
    }

    public Token nextToken() throws IOException {
        int ch;

        // Get the next non-whitespace character
        do {
            ch = getch();
            if (ch < 0)
                return Token.EOF;
        } while (Character.isWhitespace((char)ch));

        // Test for the usual operators
        if (ch == '+') return Token.PLUS;
        if (ch == '-') return Token.MINUS;
        if (ch == '*') return Token.TIMES;
        if (ch == '/') return Token.DIVIDE;
        if (ch == '^') return Token.POWER;
        if (ch == '(') return Token.LPAREN;
        if (ch == ')') return Token.RPAREN;

        if (ch == 'x' || ch == 'X')
            return Token.VARIABLE;

        if (Character.isLetter((char)ch))
            return readFunction(ch);

        if (Character.isDigit((char)ch))
            return readNumber(ch);

        throw new IOException("unknown character " + (char)ch);
    }

    private Token readFunction(int ch) throws IOException {
        StringBuffer sb = new StringBuffer(16);
        while (Character.isLetter((char)ch)) {
            sb.append((char)ch);
            ch = getch();
        }
        unget(ch);
        return new Token(sb.toString().toLowerCase());
    }

    private Token readNumber(int ch) throws IOException {
        StringBuffer sb = new StringBuffer(16);
        while (Character.isDigit((char)ch)) {
            sb.append((char)ch);
            ch = getch();
        }
        if (ch == '.') {
            sb.append('.');
            ch = getch();
            while (Character.isDigit((char)ch)) {
                sb.append((char)ch);
                ch = getch();
            }
        }
        if (ch == 'e' || ch == 'E') {
            sb.append('e');
            ch = getch();
            if (ch == '+' || ch == '-') {
                sb.append((char)ch);
                ch = getch();
            }
            while (Character.isDigit((char)ch)) {
                sb.append((char)ch);
                ch = getch();
            }
        }
        unget(ch);
        try {
            return new Token(new Double(sb.toString()));
        } catch (NumberFormatException e) {
            throw new IOException("number format exception");
        }
    }

    private int getch() throws IOException {
        if (ahead != NONE) {
            int ch = ahead;
            ahead = NONE;
            return ch;
        }
        return r.read();
    }

    private void unget(int ch) {
        if (ahead != NONE)
            throw new Error("unget called twice");
        ahead = ch;
    }

    private Reader r;
    private static final int NONE = -99;
    private int ahead = NONE;
}